/*-------------------------------------------------------------
 *  Copyright (c) 2010-2015, MSS ComBlock. All rights reserved.
 *
 * 	Filename: main.c
 *  Last modified: 12/12/15
 *  Authors:
 *  	Alain Zarembowitch / MSS ComBlock
 *  Version: 2.2
 *  Inheritance: comblock1400 Atmel code
 *  Description: main program. Includes main loop for handling commands on a polling basis.
 *
 *  Dependencies/Pre-requisite: N/A
 *
 *  Rev 2 4-24-11 AZ
 *  Editorial changes to minimize the number of warnings.
 *  Implemented independent RF and Baseband AGC loops.
*
 *  Rev 2.2 12/12/15 AZ
 *  Added usb connection/disconnection in main, depending on the presence of the USB VBUS
 -------------------------------------------------------------*/
#define __MAIN_C__

#include "main.h"
#include "uart.h"
#include "utils.h"
#include "iap.h"
#include "rf.h"

/*=========================================================================*/
/*  DEFINE: All Structures and Common Constants                            */
/*=========================================================================*/
volatile unsigned long SysTickCnt; /* SysTick Counter                    */
unsigned long cpuFreq; // CPU frequency in Hz.
volatile unsigned long timer[NTIMERS];
// timers, expressed in multiple of the system tick (10ms). 0 means expired.
// volatile because the Systick ISR will modify the values
// timer[0] = flash timer
// timer[1] = forced delay to save the config into flash
// timer[2] = inactivity timer
volatile string_struct rxLine; // rx command line
volatile string_struct txLine; // tx command line
// separate rx and tx lines (unlike previous Atmel code). Allows concurrent tx and rx and interrupts.
char address[4]; // ComBlock address, assigned dynamically (string of 3 characters + \0 end of string)
long commandPortNo; // port through which this board receives commands: 1 = top (USB ARM), 2 = right, 3 = bottom (USB FPGA), 4 = left (ComBlock rule)
unsigned char forwardMask; // Mask output ports. bit0 = top port1, bit1 = right port 2, etc. '1' means the port is blocked, '0' means the port is enabled for output
struct config_hw configHw; // hardware-related configuration.
struct config_sw configSw; // software/FPGA-related configuration volatile
char regv[NREGS]; // control registers, volatile copy of configSw.reg[]
int active_sw_config; // current active configuration (generally the same as default_sw_config, but it could be temporary different until the new FPGA config is loaded)
int usb_softconnect;	// 0 for no vbus/no connect, 1 for vbus/connected
unsigned long RF_gain1;	// receiver RF gain
unsigned long RF_gain2;	// receiver RF gain
long RF_gain_too_small_duration; // below normal range samples timer
long RF_gain_too_large_duration; // above normal range samples timer
long fc;	// estimated receiver center frequency (needed to compute attenuators response)
unsigned long rx_txnlast;	// detect changes in RX_TXN1 or RX_TXN2

/***************************************************************************/
/*  main                                                                   */
/***************************************************************************/
int main(void) {
	// local variables
	unsigned char a, tx1_power_det_MSB, tx2_power_det_MSB;
	unsigned int i, c, reg_no, s78_decode, s789_decode, sect_no;
	int x;
	int sn[4];
	struct config_sw cs; // work area for a software configuration (don't store anything of value here)

	// initial ARM processor configuration (clock frequency, system tick, etc)
	initConfig();

	// initialize UARTs 0,2,3. UART2 is reserved for future use (demodulated data in a self-standing receiver for example).
	rst: uart_init();


	/* POSSIBLE CHANGES NEXT REV
	 *
	 * 1. 	PINMODE3 = 0x30000000;
			// all pull-ups except the following:
			// pull-down on the VBUS pin
	 * otherwise VBUS pin may be staying high for a long time after a cable is removed. See IAI V5 board
	 *
	 * 2. double check if USB_CONNECT with external MOSFET is enabled. If not, copy COM-3011 code.
	 *
	 */







	// When the internal flash sector CONFIG_SECTOR is blank (all FFs), initialize it to 'all zeros'.
	initialize_configs();
	// load last-saved configuration from non-volatile (flash) memory
	read_hw_config(&configHw);
	// Only one personality in the COM-3505.
	active_sw_config = 0;

	// Semantics: we make a difference between the active_sw_config (currently loaded and running) and
	// default_sw_config which can be changed by the user prior to FPGA reboot. N/A here (one configuration)
	read_sw_config(active_sw_config, &configSw);
	memcpy(&regv, configSw.reg, sizeof(regv)); // initialize the temporary control registers with the previously saved ones.

	// ComBlock address at power up is "000"
	snprintf(&address[0],sizeof(address), "000");


	// devices initialization (one time)
	// ADC initialization
	ADC_init();
	// MAX2829 transceivers
	max2829_init(1);	// transceiver 1
	max2829_init(2);	// transceiver 2
	// initialize RF gain control (RF_gain expressed in dBs)
	RF_gain1 = (unsigned long)(regv[2]) & 0x000000FF;
	RF_gain2 = (unsigned long)(regv[5]) & 0x000000FF;
	setRxGain(RF_gain1, 1);
	setRxGain(RF_gain2, 2);
	DAC_init();	// used primarily as analog test point

	// main loop -------------------------------------------------------------------------------
	while (1) {
		// USB softconnect management based on USB VBUS presence
		USB_connect_ctrl();

		// is it time to save the current active configuration?
		// timer[1] limits how frequently the configuration can be saved,
		// timer[2] indicates persistent inactivity.
		if (((configHw.save_request_pending) || (configSw.save_request_pending))
				&& (timer[2] == 0) && (timer[1] == 0)) {
			save_configuration(-1, &cs); // save the hardware + active software configurations. no alternate configuration.
		}

		// service UARTs (polling, but optimized for smaller loop period)
		if(U0LSR & 0x00000001){		// something to read
			uart_rxpoll();
		}
		if(txLine.wptr > txLine.rptr){	// something to write
			uart_txpoll();
		}

		// service tx/rx# (polling)
		if(rx_txnlast != (FIO0PIN & 0x00018000)){
			//either RX_TXN1 or RX_TXN2 or both just changed
			rx_txn();
			rx_txnlast = FIO0PIN & 0x00018000;	// remember last value
		}

		// constantly service the ADC
		ADC_poll();

		// Rx gain control (update only while receiving)
		// Tx gain control (update only during transmission)
		if(rx_txn1)
			RxAGC_poll(1);
		else
			setTxGain(1);

		if(rx_txn2)
			RxAGC_poll(2);
		else
			setTxGain(2);

		// analog test point
		if(regv[6] & 0x0F){
			analogTP_poll();
		}

		// did we receive any message?
		// get a complete message from @ to \n (non-blocking)
		if ((rxLine.msg_complete) && (rxLine.rptr == 0)) {
			// When done processing the received message in rxLine, move the rptr = wptr to allow subsequent messages to be received.

			// re-arm inactivity timer because of new message just received.
			// We need to know when there is inactivity so that we can devote 120ms to update the configuration in flash.
			timer[2] = TIMER2;

			/*
			 // echo message (debugging)
			 memcpy(txLine.buf, rxLine.buf, rxLine.wptr);
			 txLine.wptr = rxLine.wptr;
			 rxLine.rptr = rxLine.wptr;	// release rx message
			 sendLine(4);
			 continue;
			 */

			// Dispatch the message: local processing, ignore or forward?
			if (dispatchLine()) {
				// rx message is for us: address at the start of the rx message matches (exactly or 111 broadcast address)
				delay_ns(25000); // delay 25uS before responding (otherwise slower Atmel-based ComBlocks may not be ready for a reply)

				// Remember which port is used to communicate with ComBlock control center
				// Criteria: received complete command message destined to this particular ComBlock.
				commandPortNo = rxLine.portNo;

				// decode 2-digit decimal reg_no at position 7-8 in the received string. Used by most commands.
				s78_decode = rxLine.buf[7] - '0';
				s78_decode = s78_decode * 10 + rxLine.buf[8] - '0';

				// decode 3-digit decimal sector number at position 7-9 in the received string (common to many commands)
				s789_decode = s78_decode * 10 + rxLine.buf[9] - '0';
				sect_no = s789_decode + 16; // translate ComBlock range 001-999 to internal flash block number range starting at 17.


				// command handler
				if (!memcmp((const char *)&rxLine.buf[4], "SAC", 3)) {
					// SAC command - set address
					memcpy(&address[0], (const char *)&rxLine.buf[7], 3);

				} else if (!memcmp((const char *)&rxLine.buf[4], "MFW", 3)) {
					// MFW command - Message Forwarding
					// 0 = no forwarding, 1-4: forward to port 1-4
					// 9: forward to all ports (except of course the one where the message came from)
					if (rxLine.buf[7] == '0') { // stop all forwarding
						forwardMask = 0xFF;
					} else if (rxLine.buf[7] == '1') { // forward to top port
						forwardMask = 0xFE;
					} else if (rxLine.buf[7] == '2') { // forward to right port
						forwardMask = 0xFD;
					} else if (rxLine.buf[7] == '3') { // forward to bottom port
						forwardMask = 0xFB;
					} else if (rxLine.buf[7] == '4') { // forward to left port
						forwardMask = 0xF7;
					} else if (rxLine.buf[7] == '9') { // forward to all ports, but the command port itself
						if (commandPortNo == 1) {
							forwardMask = 0xF1;
						} else if (commandPortNo == 2) {
							forwardMask = 0xF2;
						} else if (commandPortNo == 3) {
							forwardMask = 0xF4;
						} else if (commandPortNo == 4) {
							forwardMask = 0xF8;
						}
					}

				} else if (!memcmp((const char *)&rxLine.buf[4], "PNG", 3)) {
					// new PNG ping command. Echoes command back
					memcpy((char *)&(txLine.buf[0]), (const char *)&(rxLine.buf[0]), rxLine.wptr);
					txLine.buf[1] = '9';
					txLine.buf[2] = '9';
					txLine.buf[3] = '9';
					txLine.wptr = rxLine.wptr;
					sendLine(commandPortNo);

				} else if (!memcmp((const char *)&rxLine.buf[4], "GMI", 3)) {
					// GMI - Get Module Identification command
					// MID, option, revision, message port
					snprintf((char *)txLine.buf, sizeof(txLine.buf), "@999MID" ARM_FW_REV "\r\n");
					txLine.buf[13] = '0' + (unsigned char) (commandPortNo);
					txLine.wptr = strlen((const char *)txLine.buf); // set message length
					sendLine(commandPortNo); // send response through the same port the query came in

				} else if (!memcmp((const char *)&rxLine.buf[4], "SDC", 3)) {
					// SDC command - set default configuration
					// decode selected Configuration number
					c = s78_decode - 1;
					// make sure configuration pointer is within allowed boundaries
					if ((c != configHw.default_sw_config) && (c < NSWCONFIGS)) {
						// the default config changed.
						configHw.default_sw_config = c;
						configHw.save_request_pending = 1; // don't forget to save changes into non-volatile flash (cleared after saving)
						// immediate save (as this command is likely to be followed by an immediate reset)
						timer[1] = 0; // be safe: prevent another previously scheduled save from interfering.
						// Note: saving takes 120ms!! during which incoming messages are not processed
						save_configuration(-1, &cs);
						//NOTE: at this point, active and default configuration numbers are different.
					}

				} else if (!memcmp((const char *)&rxLine.buf[4], "GDC", 3)) {
					// GDC command - Get Default Configuration command
					// prepare DCF response
					c = configHw.default_sw_config + 1; // display configuration number starting at 1 (more natural for the user)
					snprintf((char *)txLine.buf, sizeof(txLine.buf), "@999DCF%02X\r\n", c);
					txLine.wptr = strlen((const char *)txLine.buf); // set message length
					sendLine(commandPortNo);

				} else if (!memcmp((const char *)&rxLine.buf[4], "SMI", 3)) {
					// SMI command - Set Module Identification

					// first check the length of the message.
					if ((rxLine.buf[12] == '\r') || (rxLine.buf[12] == '\n')) {
						// message is of the form @xxxSMI + MID + 1character option
						// thus 12 or 13 characters. use active configuration.
						// verify that the selected configuration is not protected unless
						// it is the first time this configuration is defined (all FF's)
						if ((!configSw.protection) || (configSw.mid[0] == 0xFF)) {
							memcpy(&configSw.mid[0], (const char *)&rxLine.buf[7], 5); // store MID + 1character option
							if (rxLine.buf[11] == '\r') // short message @xxxSMIYYYY\r\n
								configSw.mid[4] = ' '; // option cannot be '\r'
							configSw.mid[5] = 0; // don't know about FPGA firmware revision here
							configSw.protection = 0; // unprotected if first-time definition
							configSw.save_request_pending = 1; // don't forget to save changes into non-volatile flash (cleared after saving)
							timer[1] = TIMER1SLOW;
							// re-arm timer[1]: wait before saving the comblock configuration
							//into flash (to consolidate multiple successive changes into one sector programming operation)
						}
					} else if ((rxLine.buf[14] == '\r') || (rxLine.buf[14]
							== '\n')) {
						// message is of the form @xxxSMI+configuration number + MID + 1character option
						// thus 15 characters
						c = s78_decode - 1; // array indices start at 0, whereas configurations start at 1
						// make sure configuration pointer does not exceed array boundaries
						if (c < NSWCONFIGS) {
							read_sw_config(c, &cs); // read the software configuration of interest from flash
							// verify that the selected configuration is not protected unless it is the first time this configuration is defined (all FF's)
							if ((!cs.protection) || (cs.mid[0] == 0xFF)) {
								memcpy(&cs.mid[0], (const char *)&rxLine.buf[9], 5); // store MID + 1character option
								cs.mid[5] = 0; // don't know about FPGA firmware revision here
								cs.protection = 0; // unprotected if first-time definition
								// immediate save this non-active configuration
								timer[1] = 0; // be safe: prevent another previously scheduled save from interfering.
								// Note: saving takes 120ms!! during which incoming messages are not processed
								save_configuration(c, &cs);
							}
						}
					}

				} else if (!memcmp((const char *)&rxLine.buf[4], "SAI", 3)) {
					// SAI command - Set assembly identification
					// For security, allow only if not already programmed
					// first check if assembly id has already been programmed
					a = 0;
					/* temporarily disabled
					 for(i = 0; i < 4; i++){
					 if(aid[i] != 0)
					 a = 1;
					 }
					 */
					if (a == 0) {
						// if aid is undefined (all 0s), allow programming
						memcpy(&configHw.aid[0], (const char *)&rxLine.buf[7], 6);
						configHw.save_request_pending = 1; // don't forget to save changes into non-volatile flash (cleared after saving)
						timer[1] = TIMER1SLOW;
						// re-arm timer[1]: wait before saving the comblock configuration
						//into flash (to consolidate multiple successive changes into one sector programming operation)
					}

				} else if (!memcmp((const char *)&rxLine.buf[4], "SAS", 3)) {
					// SAS - Set Assembly Serial number,
					// For security, allow only if eeprom is not already programmed (i.e. 0000..)
					// first check if serial number has already been programmed
					a = 0;
					for (i = 0; i < sizeof(configHw.asn); i++) {
						if (configHw.asn[i] != 0)
							a = 1;
					}
					if (a == 0) {
						// if assembly serial number is null, allow programming
						memcpy(&configHw.asn[0], (const char *)&rxLine.buf[7],
								sizeof(configHw.asn));
					}

				} else if (!memcmp((const char *)&rxLine.buf[4], "GAI", 3)) {
					// GAI command - Get Assembly ID
					snprintf((char *)txLine.buf, sizeof(txLine.buf), "@999AID      \r\n"); // workaround problem with sprintf.
					memcpy((char *)&txLine.buf[7], (const char *)configHw.aid, 6);
					txLine.wptr = strlen((const char *)txLine.buf); // set message length
					sendLine(commandPortNo);

				} else if (!memcmp((const char *)&rxLine.buf[4], "GSN", 3)) {
					// GSN command - Get Serial Number
					// read unique LPC17xx device serial number
					read_sn(sn);
					snprintf((char *)txLine.buf, sizeof(txLine.buf), "@999MSN%10u %10u %10u %10u\r\n",
							sn[0], sn[1], sn[2], sn[3]); // unsigned decimal
					txLine.wptr = strlen((const char *)txLine.buf); // set message length
					sendLine(commandPortNo);

				} else if (!memcmp((const char *)&rxLine.buf[4], "GAS", 3)) {
					// GAS - Get Assembly Serial number command
					snprintf((char *)txLine.buf, sizeof(txLine.buf), "@999ASN%10s\r\n", configHw.asn);
					txLine.wptr = strlen((const char *)txLine.buf); // set message length
					sendLine(commandPortNo);

				} else if ((!memcmp((const char *)&rxLine.buf[4], "SRG", 3)) || (!memcmp(
						(const char *)&rxLine.buf[4], "SRT", 3))) {
					// SRG - Set Register (persistent/non-volatile storage)
					// SRT - Set Register Temporary (i.e. volatile, lost when powering off)

					// compute register address. Address comprises at least two digits.
					c = s78_decode;
					// check whether destination register is 2 digit or 3 digit long
					if ((rxLine.buf[11] != '\r') && (rxLine.buf[11] != '\n')) {
						// three-digit register number
						c = c * 10 + rxLine.buf[9] - '0';
						// compute register value. Value starts at char 10
						a = hex2bin((char *)&rxLine.buf[10]);
					} else {
						// two-digit register number
						// compute register value. Value starts at char 9
						a = hex2bin((char *)&rxLine.buf[9]);
					}
					reg_no = c;
					// update the configuration register
					if (reg_no < sizeof(configSw.reg)) {
						regv[reg_no] = a;
						// if SRG, also save in flash
						if (rxLine.buf[6] == 'G') {
							configSw.reg[reg_no] = a;
							configSw.save_request_pending = 1; // don't forget to save changes into non-volatile flash (cleared after saving)
							timer[1] = TIMER1SLOW;
							// re-arm timer[1]: wait before saving the comblock configuration
							//into flash (to consolidate multiple successive changes into one sector programming operation)
						}
					}
					if(reg_no == 0){
						// reg0 frequency selection for transceiver 1, power down
						program_tuner_pll(1);
						power_up_down1();
					} else if(reg_no == 1){
						// reg1. LPFs, transceiver 1, tx-only, rx-only, rx_txn mode
						bandwidth(0x01);
						rx_txn();
					} else if(reg_no == 2){
						// reg2. receiver RF gain for transceiver 1
						RF_gain1 = (long)(regv[2]) & 0x000000FF;
						setRxGain(RF_gain1, 1);	// set initial RF gain
					} else if(reg_no == 3){
						// reg3 frequency selection for transceiver 2, power down
						program_tuner_pll(2);
						power_up_down2();
					} else if(reg_no == 4){
						// reg4. LPFs, transceiver 2, tx-only, rx-only, rx_txn mode
						bandwidth(0x02);
						rx_txn();
					} else if(reg_no == 5){
						// reg5. receiver RF gain for transceiver 2
						RF_gain2 = (long)(regv[5]) & 0x000000FF;
						setRxGain(RF_gain2, 2);	// set initial RF gain
					}

				} else if (!memcmp((const char *)&rxLine.buf[4], "GSR", 3)) {
					// GSR command - Get Status Register
					// check whether 1,2 or 3 digits
					c = rxLine.buf[7] - '0';
					if ((rxLine.buf[8] != '\r') && (rxLine.buf[8] != '\n')) {
						// 2 or 3 digit register number
						c = c * 10 + rxLine.buf[8] - '0';
					}
					if ((rxLine.buf[9] != '\r') && (rxLine.buf[9] != '\n')) {
						// 3 digit register number
						c = c * 10 + rxLine.buf[9] - '0';
					}
					reg_no = c;

					// alternative status registers used for production tests (unused afterwards)
					a = 0;
					if (reg_no == 0) {
						// power check
						a = power_check();	// SREG0 bit 0
					} else	if (reg_no == 1) {
						// ch1/ch2 rf synthesizer lock status  SREG1 bits 0,1
						a = (unsigned char)(rf_synth_lock_detect());
					} else	if (reg_no == 2) {
						// ch1 RSSI. analog voltage at the ADC is 0.5 - 2.5V
						// convert to 8 bits status.
						x = rx1_rssi - 621;	// remove 0.5V pedestal
						if(x < 0)
							x = 0;
						x = (x * 33) / 20;	// restore full scale. 0.5 - 2.5V
						// clamp
						if(x > 4095)
							x = 4095;
						a = (x >> 4) & 0x00FF;
					} else	if (reg_no == 3) {
						// ch2 RSSI. analog voltage at the ADC is 0.5 - 2.5V
						// convert to 8 bits status.
						x = rx2_rssi - 621;	// remove 0.5V pedestal
						if(x < 0)
							x = 0;
						x = (x * 33) / 20;	// restore full scale. 0.5 - 2.5V
						// clamp
						if(x > 4095)
							x = 4095;
						a = (x >> 4) & 0x00FF;
					} else	if (reg_no == 4) {
						a = RF_gain1 & 0x00FF;
					} else	if (reg_no == 5) {
						a = RF_gain2 & 0x00FF;
					} else	if (reg_no == 6) {
						a = tx1_power_det & 0x00FF;	// LSB of the 12-bit ADC sample
						tx1_power_det_MSB = (tx1_power_det >> 8) & 0x00FF;	// save MSB
					} else	if (reg_no == 7) {
						a = tx1_power_det_MSB;
					} else	if (reg_no == 8) {
						a = tx2_power_det & 0x00FF;	// LSB of the 12-bit ADC sample
						tx2_power_det_MSB = (tx2_power_det >> 8) & 0x00FF;	// save MSB
					} else	if (reg_no == 9) {
						a = tx2_power_det_MSB;
					}
					snprintf((char *)txLine.buf, sizeof(txLine.buf), "@999RGV%02X\r\n", a);
					txLine.wptr = strlen((const char *)txLine.buf); // set message length
					sendLine(commandPortNo);


				} else if ((!memcmp((const char *)&rxLine.buf[4], "GRG", 3)) || (!memcmp(
						(const char *)&rxLine.buf[4], "GRT", 3))) {
					// GRG command - get register (read from flash memory)
					// GRGT command - get register temporary (read from volatile memory)

					// check whether 1,2 or 3 digits
					c = rxLine.buf[7] - '0';
					if ((rxLine.buf[8] != '\r') && (rxLine.buf[8] != '\n')) {
						// 2 or 3 digit register number
						c = c * 10 + rxLine.buf[8] - '0';
					}
					if ((rxLine.buf[9] != '\r') && (rxLine.buf[9] != '\n')) {
						// 3 digit register number
						c = c * 10 + rxLine.buf[9] - '0';
					}
					reg_no = c;

					if ((reg_no < sizeof(configSw.reg)) && (rxLine.buf[6]
							== 'G'))
						a = configSw.reg[reg_no]; // GRG, read non-volatile register
					else if ((reg_no < sizeof(regv)) && (rxLine.buf[6] == 'T')) // make sure we do not write past the buffer end
						a = regv[reg_no]; // GRT, read volatile register
					else
						// trying to read non-volatile registers past index NREG_V
						a = 0;

					snprintf((char *)txLine.buf, sizeof(txLine.buf), "@999RGV%02X\r\n", a);
					txLine.wptr = strlen((const char *)txLine.buf); // set message length
					sendLine(commandPortNo);


//				} else if (!memcmp((const char *)&rxLine.buf[4], "TST", 3)) {
//					// test test test
//					// read data at sector 17
//					c = 0x00018000 + 128 * s789_decode;
//					memcpy(&buf1[0], (void *) (c), sizeof(buf1));
//					for (i = 0; i < sizeof(buf1); i++) {
//						snprintf((char *)&txLine.buf[2 * i], 3, "%02X", buf1[i]);
//					}
//					snprintf((char *)&txLine.buf[2 * i], 3, "\r\n");
//					txLine.wptr = strlen((const char *)txLine.buf); // set message length
//					sendLine(commandPortNo);

				} else if (!memcmp((const char *)&rxLine.buf[4], "RST", 3)) {
					// RST command - reset
					goto rst;

				}

				// end of message processing section
				rxLine.rptr = rxLine.wptr; // release rx message


			}
		}
	}
	// end of main loop ------------------------------------------------------------------------


	return (0);
}

/*** EOF ***/
