/*-------------------------------------------------------------
 *  Copyright (c) 2010, MSS ComBlock. All rights reserved.
 *
 * 	Filename: usb1.c
 *  Last modified: 8/24/10
 *  Authors:
 *  	Alain Zarembowitch / MSS ComBlock
 *  Version: 1.0
 *  Inheritance: USBCORE.C provided by NXP in LPC1700CMSIS. See Keil copyright statement below.
 *  Description: USB protocol layer. Relies on usb0.c for lower-level / hardware interface.
 *
 *
 *  Dependencies/Pre-requisite: N/A
 -------------------------------------------------------------*/
/*----------------------------------------------------------------------------
 *      U S B  -  K e r n e l
 *----------------------------------------------------------------------------
 *      Name:    USBCORE.C
 *      Purpose: USB Core Module
 *      Version: V1.10
 *----------------------------------------------------------------------------
 *      This software is supplied "AS IS" without any warranties, express,
 *      implied or statutory, including but not limited to the implied
 *      warranties of fitness for purpose, satisfactory quality and
 *      noninfringement. Keil extends you a royalty-free right to reproduce
 *      and distribute executable files created using this software for use
 *      on NXP Semiconductors LPC family microcontroller devices only. Nothing
 *      else gives you the right to use this software.
 *
 *      Copyright (c) 2005-2009 Keil Software.
 *---------------------------------------------------------------------------*/

#include "usb0.h"
#include "usb1.h"
#include "usb.h"
#include "usbdesc.h"

// Global USB variables
uint16_t  USB_DeviceStatus;
uint8_t  USB_DeviceAddress;
uint8_t  USB_Configuration;
uint32_t USB_EndPointMask;
uint32_t USB_EndPointHalt;
uint8_t  USB_NumInterfaces;
uint8_t EP0Buf[USB_MAX_PACKET0];
USB_SETUP_PACKET SetupPacket;
USB_EP_DATA EP0Data;

/*
 *  USB Request - Setup Stage
 *    Parameters:      None (global SetupPacket)
 *    Return Value:    None
 */

void USB_SetupStage(void) {
	usbReadEp(0, (unsigned char *) (&SetupPacket), sizeof(SetupPacket));
}

/*
 *  USB Request - Data In Stage
 *    Parameters:      None (global EP0Data)
 *    Return Value:    None
 */

void USB_DataInStage(void) {
	uint32_t cnt;

	if (EP0Data.Count > USB_MAX_PACKET0) {
		cnt = USB_MAX_PACKET0;
	} else {
		cnt = EP0Data.Count;
	}
	usbWriteEp(1, EP0Data.pData, cnt);
	EP0Data.pData += cnt;
	EP0Data.Count -= cnt;
}

/*
 *  USB Request - Data Out Stage
 *    Parameters:      None (global EP0Data)
 *    Return Value:    None
 */

void USB_DataOutStage(void) {
	uint32_t cnt;

	cnt = usbReadEp(0, EP0Data.pData, USB_MAX_PACKET0);
	EP0Data.pData += cnt;
	EP0Data.Count -= cnt;
}

/*
 *  USB Request - Status In Stage
 *    Parameters:      None
 *    Return Value:    None
 */

void USB_StatusInStage(void) {
	usbWriteEp(1, 0, 0);
}

/*
 *  USB Request - Status Out Stage
 *    Parameters:      None
 *    Return Value:    None
 */

void USB_StatusOutStage(void) {
	usbReadEp(0, EP0Buf, USB_MAX_PACKET0);
}



// State Machine for Control Endpoint 0 (physical endpoints 0/1)
void USB_ep0(long in_outn, long epSt) {
	if ((!in_outn) && (epSt & 0x00000004)) {
		// SETUP event
		// read SETUP packet
		USB_SetupStage();
		EP0Data.Count = SetupPacket.wLength;
		switch (SetupPacket.bmRequestType.BM.Type) {

		case REQUEST_STANDARD:
			switch (SetupPacket.bRequest) {

			case USB_REQUEST_GET_STATUS:
				break;
			case USB_REQUEST_CLEAR_FEATURE:
				break;
			case USB_REQUEST_SET_ADDRESS:
				switch (SetupPacket.bmRequestType.BM.Recipient) {
				case REQUEST_TO_DEVICE:
					USB_DeviceAddress = SetupPacket.wValue.WB.L;//0x80 | SetupPacket.wValue.WB.L;
					USB_StatusInStage();
					setAddress(USB_DeviceAddress);
					break;
				default:
					break;
				}
				break;
			case USB_REQUEST_GET_DESCRIPTOR:
				if (!USB_GetDescriptor()) {
					goto stall_i;
				}
				break;
			case USB_REQUEST_SET_DESCRIPTOR:
				/*stall_o:*/
				setStallEndpoint(0);
				EP0Data.Count = 0;
				break;
			case USB_REQUEST_GET_CONFIGURATION:
				break;
			case USB_REQUEST_SET_CONFIGURATION:
				break;
			case USB_REQUEST_GET_INTERFACE:
				break;
			case USB_REQUEST_SET_INTERFACE:
				break;
			default:
				goto stall_i;
			}
			break;

		case REQUEST_CLASS:
		case REQUEST_VENDOR:
		default:
			stall_i: setStallEndpoint(1);
			EP0Data.Count = 0;
			break;

		}
	} else if ((!in_outn) && ((epSt & 0x00000006) == 0)) {
		// OUT event
	} else if (in_outn && ((epSt & 0x00000002) == 0)) {
		// IN event
	} else if ((!in_outn) && ((epSt & 0x00000006) == 2)) {
		// OUT Stalled event
		clearStallEndpoint(0);
	} else if (in_outn && ((epSt & 0x00000002) == 2)) {
		// IN Stalled event
		clearStallEndpoint(1);
	}
	return;
}

// State Machine for Bulk in/out Endpoint 2 (physical endpoints 4/5)
void USB_ep2(long rx_txn, long epSt) {
	// TODO
	return;
}

/*
 *  Get Descriptor USB Request
 *    Parameters:      None (global SetupPacket)
 *    Return Value:    TRUE - Success, FALSE - Error
 */

inline unsigned long USB_GetDescriptor(void) {
	uint8_t *pD;
	uint32_t len, n;

	switch (SetupPacket.bmRequestType.BM.Recipient) {
	case REQUEST_TO_DEVICE:
		switch (SetupPacket.wValue.WB.H) {
		case USB_DEVICE_DESCRIPTOR_TYPE:
			EP0Data.pData = (uint8_t *) USB_DeviceDescriptor;
			len = USB_DEVICE_DESC_SIZE;
			break;
		case USB_CONFIGURATION_DESCRIPTOR_TYPE:
			pD = (uint8_t *) USB_ConfigDescriptor;
			for (n = 0; n != SetupPacket.wValue.WB.L; n++) {
				if (((USB_CONFIGURATION_DESCRIPTOR *) pD)->bLength != 0) {
					pD += ((USB_CONFIGURATION_DESCRIPTOR *) pD)->wTotalLength;
				}
			}
			if (((USB_CONFIGURATION_DESCRIPTOR *) pD)->bLength == 0) {
				return 0;
			}
			EP0Data.pData = pD;
			len = ((USB_CONFIGURATION_DESCRIPTOR *) pD)->wTotalLength;
			break;
		case USB_STRING_DESCRIPTOR_TYPE:
			EP0Data.pData = (uint8_t *) USB_StringDescriptor
					+ SetupPacket.wValue.WB.L;
			len = ((USB_STRING_DESCRIPTOR *) EP0Data.pData)->bLength;
			break;
		default:
			return 0;
		}
		break;
	case REQUEST_TO_INTERFACE:
		switch (SetupPacket.wValue.WB.H) {
#if USB_HID
		case HID_HID_DESCRIPTOR_TYPE:
		if (SetupPacket.wIndex.WB.L != USB_HID_IF_NUM) {
			return 0; /* Only Single HID Interface is supported */
		}
		EP0Data.pData = (uint8_t *)USB_ConfigDescriptor + HID_DESC_OFFSET;
		len = HID_DESC_SIZE;
		break;
		case HID_REPORT_DESCRIPTOR_TYPE:
		if (SetupPacket.wIndex.WB.L != USB_HID_IF_NUM) {
			return 0; /* Only Single HID Interface is supported */
		}
		EP0Data.pData = (uint8_t *)HID_ReportDescriptor;
		len = HID_ReportDescSize;
		break;
		case HID_PHYSICAL_DESCRIPTOR_TYPE:
		return 0; /* HID Physical Descriptor is not supported */
#endif
		default:
			return 0;
		}
		break;
	default:
		return 0;
	}

	if (EP0Data.Count > len) {
		EP0Data.Count = len;
	}
	USB_DataInStage();

	return 1;
}

